#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <math.h>
#include <fortran.h>

#define MAX_N_ORBIT  10
#define LN_LEN  300
#define STR_LEN 80

const h_to_ry = 2.0;
const double pi = 3.141592653589793;
const double fourpi = 4 * 3.141592653589793;

/* forward declaration */
int skip_head(FILE *fp, int count);
int read_fhi_psp(char *potfname, int *Znuc, int* fhi_ngrid, double* fhi_clog,
		  double *fhi_r[], double *fhi_V[], double *fhi_rPsi[], 
		  double **fhi_r_core, double **fhi_n_core, char *nlcore);
double simps(int n, double* f, double h);
void printout(char *nameat, char *corr, char *irel,char *nlcore,
	      int Znuc, int nql, double delql, double Vloc0, double *Vloc,
	      int nl, int *norm, double **fql,
              double *cdc, double *cdd, int print_level);
void help();


/* Fortran routines to write data to file. */
void WRITEPOT(int *Znuc, int *nql, double *delql, double *Vloc0,
	      double *Vloc, 
	      int *nl, int *norm,
	      double *ar,
	      double *cdc, double *cdd,
              _fcd nameat, _fcd corr,
              _fcd irel, _fcd core,
              _fcd berk_pot_name);



int main(int argc, char* argv[])
{
  FILE *pot_f, *outf;
  char nameat[2], corr[2], irel[3], nlcore[4], message[130];
  int Znuc, lloc, nql, norm[MAX_N_ORBIT];
  double delql, Vloc0, *Vloc, *ar, *fql[MAX_N_ORBIT], anorm[MAX_N_ORBIT];
  double *cdc, *cdd;

  int fhi_ngrid[MAX_N_ORBIT], nl;
  double fhi_clog[MAX_N_ORBIT];
  double *fhi_r[MAX_N_ORBIT], *fhi_V[MAX_N_ORBIT], *fhi_rPsi[MAX_N_ORBIT];
  double *fhi_r_core, *fhi_n_core;
  double fill[MAX_N_ORBIT];
  
  char line[LN_LEN], berk_pot_name[STR_LEN] = "NULL";
  int i, j, l, max_ngrid, ngrid, max_fill_l, printlevel;
  double *fun, *V0, *V1, *r, *rPsi, clog, q, arg;


  if (argc < 2) {
    fprintf(stderr,"Usage: %s <fhi-pseudopotential> [options]\n",
            argv[0]);
    fprintf(stderr,"transpot -h   for list of options.\n");
    /* "<l_loc> <nql> <delql>\n", */
    fflush(stderr);
    exit(1);
  }

  /* setting defaults */
  lloc = 2;  /* default using d as local */
  nql  = 1601; /* default using 1601 points */
  delql = 0.015;
  for (i = 0; i < MAX_N_ORBIT; i++) {
    fill[i] = 0.0;
  }
  fill[0] = 2.0;  /* default, fill s with 2, p with 2 */
  fill[1] = 2.0; 
  strcpy(nameat,"NN");
  strcpy(corr,"ca");
  strcpy(irel,"nrl");
  strcpy(nlcore,"core");
  printlevel = 0;

  for (i = 1; i < argc; i++) {
    if (strncmp(argv[i],"-h",2)==0) {
      help();
      exit(0);
    } else if (strcmp(argv[i],"-loc")==0) {
      i++;
      lloc = atoi(argv[i]);
    } else if (strcmp(argv[i],"-nql")==0) {
      i++;
      nql = atoi(argv[i]);
    } else if (strcmp(argv[i],"-dql")==0) {
      i++;
      delql = atof(argv[i]);
    } else if (strcmp(argv[i],"-fill")==0) {
      i++;
      l = atoi(argv[i]);
      i++;
      fill[l] = atof(argv[i]);
    } else if (strcmp(argv[i],"-name")==0) {
      i++;
      strncpy(nameat,argv[i],2);
    } else if (strcmp(argv[i],"-corr")==0) {
      i++;
      strncpy(corr,argv[i],2);
    } else if (strcmp(argv[i],"-rel")==0) {
      i++;
      strncpy(irel,argv[i],3);
    } else if (strcmp(argv[i],"-core")==0) {
      i++;
      strncpy(nlcore,argv[i],4);
    } else if (strcmp(argv[i],"-print")==0) {
      i++;
      printlevel = atoi(argv[i]);
    } else if (strcmp(argv[i],"-o")==0) {
      i++;
      strcpy(berk_pot_name,argv[i]);
    }
  }
  if (strcmp(berk_pot_name,"NULL")==0) {
    sprintf(berk_pot_name,"%s_POT.DAT",argv[1]);
  }
  fprintf(stdout,"Generate paratec pot with:\n");
  fprintf(stdout,"loc = %d, nql = %d, delql = %g\n",
	  lloc, nql, delql);
  for (i= MAX_N_ORBIT-1; i>-1; i--) {
    if (fill[i] != 0.0) {
      max_fill_l = i;
      break;
    }
  }
  fprintf(stdout,"fill (%d): ", max_fill_l);
  for (i=0; i <= max_fill_l; i++) {
    fprintf(stdout," %g", fill[i]);
  }
  fprintf(stdout,"\n");


  /*
   * Read in the FHI pseudopotential
   */
  nl = read_fhi_psp(argv[1], &Znuc, fhi_ngrid, fhi_clog, 
		    fhi_r, fhi_V, fhi_rPsi, &fhi_r_core, &fhi_n_core,
		    nlcore);

  fprintf(stdout,"Finish reading. Start converting...\n");
  fflush(stdout);

  /*
  fprintf(stdout,">> Zv = %d\n", Znuc);
  for (i = 0; i < nl; i++) {
    fprintf(stdout,">> l = %d, NR = %d, Clog = %g\n", 
	    i, fhi_ngrid[i], fhi_clog[i]);
    for (j = 0; j < fhi_ngrid[i]; j++) {
      fprintf(stdout, ">> %6d %20.10e %20.10e %20.10e\n",
	      j, fhi_r[i][j], fhi_rPsi[i][j], fhi_V[i][j]);
    }
  }
  for (i = 0; i < fhi_ngrid[0]; i++) {
    fprintf(stdout,">> %20.10e %20.10e\n", fhi_r_core[i], fhi_n_core[i]);
  }
  */

  /* allocate temporary array for fourier transform */
  for (i = 0, max_ngrid = fhi_ngrid[0]; i < nl; i++) {
    if (max_ngrid < fhi_ngrid[i])
      max_ngrid = fhi_ngrid[i];
  }
  max_ngrid += 20;  /* just to be safe */
  fun = (double*)malloc(sizeof(double)*max_ngrid);
  


  /*=====================================================================*
   *                                                                     *
   * For local pseudopotential                                           *
   *                                                                     *
   *=====================================================================*/

  fprintf(stdout,"Begin local potential...\n"); fflush(stdout);

  /* calculate V(G=0) as the difference from columb */
  ngrid = fhi_ngrid[lloc];
  clog  = fhi_clog[lloc];
  V0    = fhi_V[lloc];
  r     = fhi_r[lloc];
  for (i = 0; i < ngrid; i++) {
    fun[i] = (V0[i] * r[i] + Znuc) * r[i] * r[i];
  }
  Vloc0 = simps(ngrid,fun,clog);
  Vloc0 *= fourpi;

  fprintf(stdout,"Begin local potential...\n"); fflush(stdout);

  /* calculate the rest Vloc(G) */
  Vloc = (double*)malloc(sizeof(double)*nql);
  for (j = 0; j < nql; j++) {
    q = (j+1)*delql;

    for (i = 0; i < ngrid; i++) {
      arg = q * r[i];
      fun[i] = (V0[i] * r[i] + Znuc) * r[i] * r[i] * sin(arg)/arg;
    }
    Vloc[j] = simps(ngrid,fun,clog);
    Vloc[j] -= Znuc/(q*q);
    Vloc[j] *= fourpi;
  }

  /* finally, FHI pseudopotential use natural unit -- Hartree, but 
     berkeley ones use Ry, a factor of 2. */
  Vloc0 *= h_to_ry;
  for (j = 0; j < nql; j++)
    Vloc[j] *= h_to_ry;


  fprintf(stdout,"Finish local potential...\n"); fflush(stdout);

  /*=====================================================================*
   *                                                                     *
   * For Core charge density                                             *
   *                                                                     *
   *=====================================================================*/

  /* fourier tranform the core charge density */
  cdc = (double*)malloc(sizeof(double)*nql);
  if (strncmp(nlcore,"nc  ",4)!=0) {
    ngrid = fhi_ngrid[0];
    clog  = fhi_clog[0];
    r     = fhi_r_core;
    for (j = 0; j < nql; j++) {
      q = (j+1)*delql;

      for (i = 0; i < ngrid; i++) {
	arg = q * r[i];
	fun[i] = fhi_n_core[i] * (r[i]*r[i]*r[i]) * sin(arg)/arg;
      }
      cdc[j] = simps(ngrid,fun,clog) * fourpi;
    }
  } else {
    for (j=0; j < nql; j++) {
      cdc[j] = 0.0;
    }
  }


  fprintf(stdout,"Finish cdc...\n");

  /*=====================================================================*
   *                                                                     *
   * For Valence charge density                                          *
   *                                                                     *
   *=====================================================================*/

  /* calculate the valence charge density */
  /* you need to tell me the fillings */
  cdd = (double*)malloc(sizeof(double)*nql);
  ngrid = fhi_ngrid[0];
  clog  = fhi_clog[0];
  r     = fhi_r[0];
  for (j = 0; j < nql; j++) {
    q = (j+1)*delql;

    for (i = 0; i < ngrid; i++) {
      arg = 0;
      for (l = 0; l <= max_fill_l; l++) {
	arg += fill[l] * fhi_rPsi[l][i] * fhi_rPsi[l][i];
      }
      fun[i] = arg * sin(q*r[i]);
    }
    cdd[j] = simps(ngrid,fun,clog)/q;
  }


  fprintf(stdout,"Finish cdd...\n");

  /*=====================================================================*
   *                                                                     *
   * For KB non-local pseudopotential                                    *
   *                                                                     *
   *=====================================================================*/

  /* generate norms:  <Psi_l| dVnl |Psi_l> */
  for (l = 0; l< nl; l++) {
    if (l == lloc) {
      norm[l] = 0;
      anorm[l] = 1.0;
      fprintf(stdout,"l = %d is local\n",l);
    } else {
      ngrid = fhi_ngrid[l];
      clog  = fhi_clog[l];
      V1    = fhi_V[l];
      rPsi  = fhi_rPsi[l];
      r     = fhi_r[l];
      for (i = 0; i < ngrid; i++) {
	fun[i] = rPsi[i]*rPsi[i]*r[i]*(V1[i]-V0[i]);
      }
      anorm[l] = simps(ngrid,fun,clog);
      fprintf(stdout,"l = %d, anorm = %f\n", l, anorm[l]);
      if (anorm[l] > 0) {
	norm[l] = 1;
	anorm[l] = sqrt(anorm[l]);
      } else if (anorm[l] < 0) {
	norm[l] = -1;
	anorm[l] = sqrt(-anorm[l]);
      }
      
      /* because of the energy unit mismatch mentioned earlier: hartree vs. Ry */
      anorm[l] *= sqrt(2.0);
    }
  }

  fprintf(stdout,"I Finish <Psi|dV|Psi>...\n");
  fflush(stdout);

  /* Generate ar */
  ar = (double*)malloc(sizeof(double)*(nql+1)*nl);
  for (l = 0; l < nl; l++) {
    fql[l] = ar + l*(nql+1);
    if (l == lloc) {
      fprintf(stdout,"found local\n");
      for (j = 0; j <= nql; j++) {
	fql[l][j] = 0.0;
      }
    } else {
      ngrid = fhi_ngrid[l];
      clog  = fhi_clog[l];
      V1    = fhi_V[l];
      rPsi  = fhi_rPsi[l];
      r     = fhi_r[l];
      fprintf(stdout,"do non-local %d\n", l);
      anorm[l] = anorm[l] / sqrt( fourpi * (2*l+1) );
      for (j = 0; j <= nql; j++) {
	q = j*delql;
	for (i = 0; i < ngrid; i++) {
	  fun[i] = rPsi[i]*r[i]*r[i]*(V1[i]-V0[i]);
	  arg = q * r[i];
	  switch (l) {
	  case 0:  /* bessel_0 = sin(x)/x; */
	    if (arg < 1e-8) {
	      arg = 1.0;
	    } else {
	      arg = sin(arg)/arg;
	    }
	    break;
	  case 1:  /* bessel_1 = (sin(x)/x-cos(x))/x */
	    if (arg < 1e-8) {
	      arg = arg*arg/3.0;
	    } else {
	      arg = (sin(arg)/arg-cos(arg))/arg;
	    }
	    break;
	  case 2:  /* bessel_2 = ((3/x^3-1/x)*sin(x)-(3/x^2)*cos(x)) */
	    if (arg < 1e-8) {
	      arg = arg*arg/15;
	    } else {
	      arg = ((3/(arg*arg)-1)*sin(arg)-3*cos(arg)/arg)/arg;
	    }
	    break;
	  default:
	    fprintf(stderr, "Error: no bessel functions for l = %d\n",l);
	    exit(-10);
	  }
	  fun[i] *= arg;
	}
	fql[l][j] = simps(ngrid,fun,clog) / anorm[l];

	/* because of the energy unit mismatch mentioned earlier: hartree vs. Ry */
	fql[l][j] *= 2.0;
      }
    }
  }

  fprintf(stdout,"Finish ar[][]...\n");

  /*=====================================================================*
   *                                                                     *
   * Ready to dump                                                       *
   *                                                                     *
   *=====================================================================*/
  printout(nameat,corr,irel,nlcore,
	   Znuc, nql, delql, Vloc0, Vloc,
	   nl, norm, fql, cdc, cdd, printlevel);
  
  fprintf(stdout,"\n\nWrite to file: %s\n", berk_pot_name);
  fflush(stdout);

  /* also convert character strings to fortran format */
  WRITEPOT(&Znuc,&nql,&delql,&Vloc0,
           Vloc, &nl, norm, ar, cdc, cdd,
           _cptofcd(nameat,2), 
	   _cptofcd(corr,2), 
	   _cptofcd(irel,3), 
	   _cptofcd(nlcore,4),
           _cptofcd(berk_pot_name,80) );


  return 0;
}


/*============================================================================*
 *                                                                            *
 * Read Pseudo potential information from file                                *
 *                                                                            *
 * 2. The berlin format from package  FHI98MD                                 *
 *    l_max indicates which angular momentum components are used              *
 *          l_max = 0,  s                                                     *
 *                  1,  s, p                                                  *
 *                  2,  s, p, d                                               *
 *    l_loc indicates which one is being used as the                          *
 *          local component.                                                  *
 *                                                                            *
 *    The file format is:                                                     *
 *     >>>>>>>( Start )>>>>>>>>>>                                             *
 *        Zv  L_max                                                           *
 *          (10 lines of number for fitted psp                                *
 *           not used here.)                                                  *
 *        ngrid  clog       #  (for each angular momentum)                    *
 *          i  r(i)  rPsi(i)  Vion(i)                                         *
 *          ...                                                               *
 *          ...                                                               *
 *        (ngrid of core electron density for                                 *
 *         nonlinear core correction)                                         *
 *          r(i) 4pi*rho(i) 4pi*rho'(i) 4pi*rho''(i)                          *
 *     <<<<<<<(  End  )<<<<<<<<<<                                             *
 *============================================================================*/
int read_fhi_psp(char *potfname, int *Znuc, int* fhi_ngrid, double* fhi_clog,
		  double *fhi_r[], double *fhi_V[], double *fhi_rPsi[], 
		  double **fhi_r_core, double **fhi_n_core, char * nlcore)
{
  FILE * potfile;
  int nl, l, j, i, n;
  double zv, ofpi = 0.25/pi;
  char line[LN_LEN];

  if ( (potfile = fopen(potfname,"r")) == NULL ) {
    fprintf(stderr, "Can't open potential file %s\n", potfname);
    exit(-1);
  }

  fprintf(stdout, "Reading potential file %s\n", potfname);
  /*
   * The first line has two numbers:
   *   Zv   number_of_angular_momentum_components
   */
  fgets(line,LN_LEN,potfile);
  sscanf(line,"%lg %d", &zv, &nl);
  *Znuc = (int)(zv);
  fprintf(stdout,"Zv = %d, Nl = %d\n", *Znuc, nl); fflush(stdout);

  /* there are 10 lines of data for pseudopotentials that 
   * are fitted to expressions. Skip. */
  for (n = 0; n < 10; n++) 
    fgets(line,LN_LEN,potfile);

  /*
   * Read in all wavefunctions and potentials.
   * All are in real space with logrithmic grid spacing.
   */
  for (l = 0; l < nl; l++) {
    fgets(line,LN_LEN,potfile);
    sscanf(line, "%d %lg", &fhi_ngrid[l], &fhi_clog[l]);

    /* clog[l] == R[l][1]/R[l][0]  is the scale of the log grid spacing. 
     * now clog[l] = D log(R), used for integration later. */
    fhi_clog[l] = log(fhi_clog[l]);

    fprintf(stdout, " l = %d:  ngrid = %d, clog = %f\n",
	    l, fhi_ngrid[l], fhi_clog[l]);
    fflush(stdout);

    fhi_r[l]   = (double*)malloc(sizeof(double)*fhi_ngrid[l]);
    fhi_rPsi[l] = (double*)malloc(sizeof(double)*fhi_ngrid[l]);
    fhi_V[l]   = (double*)malloc(sizeof(double)*fhi_ngrid[l]);

    for (j = 0; j < fhi_ngrid[l]; j++) {
      fgets(line,LN_LEN,potfile);
      sscanf(line,"%*d %lg %lg %lg",
	     &(fhi_r[l][j]), 
	     &(fhi_rPsi[l][j]), 
	     &(fhi_V[l][j]));
    }

  }
  fflush(stdout);

  /*
   * Read core electron density, for nonlinear core correction.
   * It doesn't have a number for its number of grid points,
   * use that of the s-potential. 'fhi_ngrid[0]'
   */
  *fhi_r_core = (double*)malloc(sizeof(double)*fhi_ngrid[0]);
  *fhi_n_core = (double*)malloc(sizeof(double)*fhi_ngrid[0]);
  for (i=0; i < fhi_ngrid[0]; i++) {
    (*fhi_n_core)[i] = 0.0;
    (*fhi_r_core)[i] = 0.0;
  }
  for (j = 0; j < fhi_ngrid[0]; j++) {
    if ( fgets(line,LN_LEN,potfile) == NULL ) {
      for (i=0; i < fhi_ngrid[0]; i++) {
	(*fhi_n_core)[i] = 0.0;
	(*fhi_r_core)[i] = -1.0;
      }
      fprintf(stdout,"no core found\n");
      strcpy(nlcore,"nc  ");
      break;
    }
    if (sscanf(line,"%lg %lg", &((*fhi_r_core)[j]), &((*fhi_n_core)[j])) != 2) {
      for (i=0; i < fhi_ngrid[0]; i++) {
	(*fhi_n_core)[i] = 0.0;
	(*fhi_r_core)[i] = -1.0;
      }
      fprintf(stdout,"no core found\n");
      strcpy(nlcore,"nc  ");
      break;
    }
    (*fhi_n_core)[j] *= ofpi;
  }

  return nl;
}


/*============================================================================*
 *                                                                            *
 * simps  integration for  f  with step length  h  for  n points              *
 * n  should be odd.                                                          *
 *                                                                            *
 *============================================================================*/
double simps(int n, double* f, double h)
{
  double result;
  int nm12 = (n-1)/2, i;

  if ( (nm12*2) != (n-1) ) {
    fprintf(stderr,"Function simps: n must be odd, but is not.");
    exit(-9);
  }

  result = f[0];
  for (i = 1; i < (n-1); i += 2)
    result += 4 * f[i] + 2 * f[i+1];
  result -= f[n-1];
  /* the weights are basically:  1 4 2 4 2 4 2 ... 4 2 4 1 */

  result *= (h/3);
  
  return(result);
}


/*============================================================================*
 *                                                                            *
 * print out some info for checking                                           *
 *                                                                            *
 *============================================================================*/
void printout(char *nameat, char *corr, char *irel,char *nlcore,
	      int Znuc, int nql, double delql, double Vloc0, double *Vloc,
	      int nl, int *norm, double **fql,
              double *cdc, double *cdd, int print_level)
{
  int i, j;
  fprintf(stdout,"Species = %s, corr = %s, irel = %s, nlcore = %s\n",
	  nameat, corr, irel, nlcore);
  fprintf(stdout,"Zv = %d, Nql = %d, dql = %g, Vloc0 = %g\n",
	  Znuc, nql, delql, Vloc0);
  if (print_level > 0) {
    fprintf(stdout,"Vloc = :\n");
    for (i = 0; i < nql; i++) {
      fprintf(stdout," %5d\t%g\n",i,Vloc[i]);
    }
  }
  fprintf(stdout,"N_orb = %d\t", nl);
  for (i = 0; i < nl; i++) 
    fprintf(stdout," %d", norm[i]);
  if (print_level > 0) {
    fprintf(stdout,"\nFql = :\n");
    for (i = 0; i < nql+1; i++) {
      for (j = 0; j < nl; j++) {
        fprintf(stdout," %20.10f",fql[j][i]);
      }
      fprintf(stdout,"\n");
    }
  }
  if (print_level > 0) {
    fprintf(stdout,"charge density (core/valence):\n");
    for (i = 0; i < nql; i++) {
      fprintf(stdout," %20.10f\t%20.10f\n", cdc[i], cdd[i]);
    }
  }

}


/*============================================================================*
 *                                                                            *
 * print out some help info                                                   *
 *                                                                            *
 *============================================================================*/
void help()
{
  fprintf(stdout,"Usage: transpot <fhi-pseudopotential> [options]\n");
  fprintf(stdout,"  options:\n");
  fprintf(stdout,"\t-h\thelp on options.\n");
  fprintf(stdout,"\t-loc [number]\tset local components, default: 2\n");
  fprintf(stdout,"\t-nql [number]\tset number of q-points, default: 1601\n");
  fprintf(stdout,"\t-dql [number]\tset dq, default: 0.015\n");
  fprintf(stdout,"\t-fill [l] [filling]\tset fillings, can be used repeatly.\n");
  fprintf(stdout,"\t\t\tdefault: f_s = 2.0, f_p = 2.0\n");
  fprintf(stdout,"\t-name [species name]\tset 2-char species name\n");
  fprintf(stdout,"\t-corr [string]\tset 2-char correlation name, default: ca\n");
  fprintf(stdout,"\t-rel [string]\tset 3-char relativistic correction, default: nrl\n");
  fprintf(stdout,"\t-core [string]\tset 4-char core correction, default: \"nc  \"\n");
  fprintf(stdout,"\t-print [number]\tset printlevel, default: 0\n");
  
}
